import { Manager, Swipe, Tap } from 'hammerjs';
import { MessagingSystem } from '../eventing/messaging-system';
import { ItemFocusEvent } from '../eventing/events';

export class DOMNavigator extends MessagingSystem {
  private currentIndex: number;
  private currentElement: HTMLElement;

  constructor(private rootElement: HTMLElement) {
    super();
    this.currentIndex = 0;
    this.currentElement = rootElement;

    this.setupGestures();
  }

  private setupGestures(): void {
    const manager = new Manager(this.rootElement);

    const swipe = new Swipe();
    const tap = new Tap({ event: 'doubletap', taps: 2 });

    manager.add(swipe);
    manager.add(tap);

    manager.on('swipe', (event) => {
      switch (event.direction) {
        case 2: // Left
          this.navigateSiblings(-1);
          break;
        case 4: // Right
          this.navigateSiblings(1);
          break;
        case 8: // Up
          this.navigateParent();
          break;
        case 16: // Down
          this.navigateChildren();
          break;
      }
    });

    manager.on('doubletap', () => {
      this.currentElement.click();
    });

    this.rootElement.addEventListener("keydown", e => {
      switch (e.key) {
        case "ArrowRight":
          this.navigateSiblings(1);
          break;
        case "ArrowLeft":
          this.navigateSiblings(-1);
          break;
        case "ArrowUp":
          this.navigateParent();
          break;
        case "ArrowDown":
          this.navigateChildren();
          break;
        case "Enter":
          this.currentElement.click();
          break;
      }
    })
  }


  private navigateSiblings(direction: number): void {
    const parent = this.currentElement.parentElement;

    if (!parent) return;

    const siblings = Array.from(parent.children);
    const newIndex = this.currentIndex + direction;

    if (newIndex >= 0 && newIndex < siblings.length) {
      this.currentIndex = newIndex;
      this.currentElement = siblings[newIndex] as HTMLElement;
    }
    this.handleFocus();
  }

  private navigateParent2(): void {
    const parent = this.currentElement.parentElement;

    if (!parent || parent === this.rootElement) return;

    this.currentElement = parent;
    this.currentIndex = Array.from(parent.parentElement.children).indexOf(parent);
    this.handleFocus();
  }

  private navigateChildren2(): void {
    if (!this.currentElement.children.length) return;

    this.currentElement = this.currentElement.children[0] as HTMLElement;
    this.currentIndex = 0;
    this.handleFocus();
  }

  private handleFocus() {
    let element = this.currentElement;
    if (this.isSingleWrapper(element)) {
      element = this.currentElement.children[0] as HTMLElement;
    }
    element.focus();
    this.sendMessage<ItemFocusEvent>({
      type: "ItemFocusEvent",
      data: {
        node: element,
        friendlyName: `${this.getFriendlyName(element)} ${this.getFriendlyTypeName(element)}`
      }
    });
  }

  private isContainer(element: HTMLElement): boolean {
    return ['div', 'label'].includes(element.tagName.toLowerCase());
  }

  private isSingleDivWrapper(element: HTMLElement): boolean {
    return this.isContainer(element) && element.children.length === 1 && this.isContainer(element.children[0] as HTMLElement);
  }

  private isSingleWrapper(element: HTMLElement): boolean {
    return this.isContainer(element) && element.children.length === 1 && !this.isContainer(element.children[0] as HTMLElement);
  }

  private navigateParent(): void {
    let currentParent = this.currentElement.parentElement;

    while (currentParent && currentParent !== this.rootElement) {
      if (!this.isSingleDivWrapper(currentParent)) {
        this.currentElement = currentParent;
        this.currentIndex = Array.from(currentParent.parentElement.children).indexOf(currentParent);
        break;
      }
      currentParent = currentParent.parentElement;
    }
    this.handleFocus();
  }

  private navigateChildren(): void {
    const children = Array.from(this.currentElement.children);
    let targetChild = children[0] as HTMLElement;

    while (targetChild && this.isSingleDivWrapper(targetChild)) {
      targetChild = targetChild.children[0] as HTMLElement;
    }

    if (targetChild) {
      this.currentElement = targetChild;
      this.currentIndex = children.indexOf(targetChild);
      this.handleFocus();
    }
  }

  public hasSiblings(): boolean {
    const parent = this.currentElement.parentElement;

    if (!parent) return false;

    const siblings = Array.from(parent.children);
    return siblings.length > 1 && (this.isContainer(this.currentElement));
  }

  private getFriendlyName(element: HTMLElement) {
    if (!element) return '';

    // Check if the element is a child of a label
    let parentLabel: HTMLLabelElement | null = null;
    let currentParent = element.parentElement;
    while (currentParent) {
      if (currentParent.tagName.toLowerCase() === 'label') {
        parentLabel = currentParent as HTMLLabelElement;
        break;
      }
      currentParent = currentParent.parentElement;
    }

    // Check if the element has a label with the 'for' or 'aria-labelledby' attribute
    const labelledById = element.getAttribute('aria-labelledby');
    const labelledBy = labelledById ? document.getElementById(labelledById) : null;

    if (element.tagName.toLowerCase() === 'div') {
      const ariaLabel = element.getAttribute('aria-label');
      const legend = element.querySelector('legend');
      const groupName = ariaLabel || (legend ? legend.textContent : '');

      if (groupName) {
        return groupName;
      }
    }

    if (parentLabel) {
      return parentLabel.textContent?.trim() || '';
    }

    if (labelledBy) {
      return labelledBy.textContent?.trim() || '';
    }

    // Check for aria-label attribute
    const ariaLabel = element.getAttribute("aria-label");
    if (ariaLabel) {
      return ariaLabel;
    }

    // Check for title attribute
    const title = element.getAttribute("title");
    if (title) {
      return title;
    }

    // Check if the element is a fieldset with a legend child
    if (element.tagName.toLowerCase() === "fieldset") {
      const legend = element.querySelector("legend");
      if (legend) {
        return legend.textContent || null;
      }
    }

    // Return the text content of the element, but only if it's not a container (e.g., div)
    if (element.tagName.toLowerCase() !== "div") {
      return element.textContent || null;
    }

    // If none of the above conditions are met, return null
    return "Group";
  }

  private getFriendlyTypeName(element: HTMLElement) {
    const tagName = element.tagName.toLowerCase();
    const inputType = (element as HTMLInputElement).type?.toLowerCase();

    switch (tagName) {
      case 'button':
        return 'Button';

      case 'input':
        if (inputType === 'checkbox') {
          const checked = (element as HTMLInputElement).checked;
          return `Checkbox (${checked ? 'Checked' : 'Unchecked'})`;
        }
        if (inputType === 'radio') {
          const checked = (element as HTMLInputElement).checked;
          return `Radio Button (${checked ? 'Selected' : 'Unselected'})`;
        }
        return 'Input';

      case 'select':
        return 'Select';

      case 'textarea':
        return 'Text Area';

      case 'li':
        return 'List item';

      case 'ul':
        return 'List';

      case 'span':
      case "p":
        return '';

      default:
        return tagName.charAt(0).toUpperCase() + tagName.slice(1);
    }

  }
}