export interface CustomKeyboardOptions {
    leftKeys: string[][];
    rightKeys: string[][];
}

export class CustomKeyboard {
    private container: HTMLElement;
    private options: CustomKeyboardOptions;
    private keys: Map<string, HTMLElement> = new Map();
    private pressedKeys: Map<number, { element: HTMLElement; key: string }> = new Map();

    constructor(container: HTMLElement, options: CustomKeyboardOptions) {
        this.container = container;
        this.options = options;
        this.createKeyboard();
        this.attachEvents();
    }

    private createKeyboard(): void {
        const leftSection = document.createElement('div');
        leftSection.classList.add('keyboard-section', 'left');
        this.createKeys(leftSection, this.options.leftKeys);

        const rightSection = document.createElement('div');
        rightSection.classList.add('keyboard-section', 'right');
        this.createKeys(rightSection, this.options.rightKeys);

        this.container.appendChild(leftSection);
        this.container.appendChild(rightSection);
    }

    private createKeys(section: HTMLElement, keysLayout: string[][]): void {
        keysLayout.forEach((rowKeys) => {
            const row = document.createElement('div');
            row.classList.add('keyboard-row');
            rowKeys.forEach((key) => {
                const keyElement = document.createElement('button');
                keyElement.textContent = key;
                keyElement.classList.add('keyboard-key');
                keyElement.dataset.key = key;
                row.appendChild(keyElement);
                this.keys.set(key, keyElement);
            });
            section.appendChild(row);
        });
    }

    private attachEvents(): void {
        // this.container.addEventListener('touchstart', this.handleTouchStart.bind(this), { passive: false });
        // this.container.addEventListener('touchend', this.handleTouchEnd.bind(this), { passive: false });
        // this.container.addEventListener('touchmove', this.handleTouchMove.bind(this), { passive: false });
        this.container.addEventListener('pointerdown', this.handlePointerDown.bind(this), { passive: false });
        this.container.addEventListener('pointerup', this.handlePointerUp.bind(this), { passive: false });
        this.container.addEventListener('pointermove', this.handlePointerMove.bind(this), { passive: false });
        this.container.addEventListener('contextmenu', this.handleContextMenu.bind(this), { passive: false });
        // this.container.addEventListener('click', this.handleClick.bind(this), { passive: false });
    }

    private handleTouchStart(event: TouchEvent): void {
        event.preventDefault();
        const target = event.target as HTMLElement;
        const key = target.dataset.key;
        if (key) {
            const touch = event.changedTouches[0];
            this.pressedKeys.set(touch.identifier, { element: target, key });
            this.keys.get(key)?.classList.add('pressed');
            this.emit('keyDown', key);
        }
    }

    private handleTouchEnd(event: TouchEvent): void {
        event.preventDefault();
        const touch = event.changedTouches[0];
        const target = document.elementFromPoint(touch.clientX, touch.clientY) as HTMLElement;
        const key = target.dataset.key;
        const pressedKeyData = this.pressedKeys.get(touch.identifier);

        if (pressedKeyData) {
            // Remove the 'pressed' class from the original key pressed
            pressedKeyData.element.classList.remove('pressed');
            this.pressedKeys.delete(touch.identifier);

            // Emit the 'keyUp' event for the initially pressed key
            this.emit('keyUp', pressedKeyData.key);

            // If the finger is released on a different key, emit the 'keyUp' event for that key
            if (key && pressedKeyData.key !== key) {
                this.emit('keyUp', key);
            }
        }
    }

    private handleTouchMove(event: TouchEvent): void {
        event.preventDefault();
        const touch = event.changedTouches[0];
        const target = document.elementFromPoint(touch.clientX, touch.clientY) as HTMLElement;
        const key = target.dataset.key;

        const pressedKeyData = this.pressedKeys.get(touch.identifier);

        if (pressedKeyData) {
            // If the finger slides over a different key, emit the 'keyDown' event for that key
            // and the 'keyUp' event for the previous key
            if (key && pressedKeyData.key !== key) {
                // Remove the 'pressed' class from the previous key and update the pressedKeys Map
                pressedKeyData.element.classList.remove('pressed');
                this.pressedKeys.set(touch.identifier, { element: target, key });

                // Add the 'pressed' class to the new key
                this.keys.get(key)?.classList.add('pressed');

                // Emit the 'keyDown' event for the new key
                this.emit('keyDown', key);

                // Emit the 'keyUp' event for the previous key
                // this.emit('keyUp', pressedKeyData.key);
            }
        }
    }

    private handleClick(event: MouseEvent) {

        event.preventDefault();
        const target = event.target as HTMLElement;
        const key = target.dataset.key;

        if (key) {
            this.keys.get(key)?.classList.add('pressed');

            // Emit the 'keyUp' event for the clicked key
            this.emit('keyUp', key);

            // Remove the 'pressed' class after a short delay
            setTimeout(() => {
                this.keys.get(key)?.classList.remove('pressed');
            }, 100);
        }
    }

    private handlePointerDown(event: PointerEvent): void {
        event.preventDefault();
        const target = event.target as HTMLElement;
        const key = target.dataset.key;
        if (key) {
            this.pressedKeys.set(event.pointerId, { element: target, key });
            this.keys.get(key)?.classList.add('pressed');
            this.emit('keyDown', key);
        }
    }

    private handlePointerUp(event: PointerEvent): void {
        event.preventDefault();
        const target = document.elementFromPoint(event.clientX, event.clientY) as HTMLElement;
        const key = target.dataset.key;

        const pressedKeyData = this.pressedKeys.get(event.pointerId);

        if (pressedKeyData) {
            // Remove the 'pressed' class from the original key pressed
            pressedKeyData.element.classList.remove('pressed');
            this.pressedKeys.delete(event.pointerId);

            // Emit the 'keyUp' event for the initially pressed key
            this.emit('keyUp', pressedKeyData.key);

            // If the finger is released on a different key, emit the 'keyUp' event for that key
            if (key && pressedKeyData.key !== key) {
                this.emit('keyUp', key);
            }
        }
    }

    private handlePointerMove(event: PointerEvent): void {
        event.preventDefault();
        const target = document.elementFromPoint(event.clientX, event.clientY) as HTMLElement;
        const key = target.dataset.key;

        const pressedKeyData = this.pressedKeys.get(event.pointerId);

        if (pressedKeyData) {
            // If the finger slides over a different key, emit the 'keyDown' event for that key
            // and the 'keyUp' event for the previous key
            if (key && pressedKeyData.key !== key) {
                // Remove the 'pressed' class from the previous key and update the pressedKeys Map
                pressedKeyData.element.classList.remove('pressed');
                this.pressedKeys.set(event.pointerId, { element: target, key });

                // Add the 'pressed' class to the new key
                this.keys.get(key)?.classList.add('pressed');

                // Emit the 'keyDown' event for the new key
                this.emit('keyDown', key);

                // Emit the 'keyUp' event for the previous key
                // this.emit('keyUp', pressedKeyData.key);
            }
        }
    }

    private handleContextMenu(event: MouseEvent): void {
        event.preventDefault();
    }

    private emit(eventName: string, key: string): void {
        const event = new CustomEvent(eventName, { detail: { key } });
        this.container.dispatchEvent(event);
    }

    on(eventName: any, callback: (event: CustomEvent) => void): void {
        this.container.addEventListener(eventName, callback);
    }

    off(eventName: any, callback: (event: CustomEvent) => void): void {
        this.container.removeEventListener(eventName, callback);
    }

    show(): void {
        this.container.classList.add('visible');
    }

    hide(): void {
        this.container.classList.remove('visible');
    }
}
